﻿Imports Infragistics.Olap
Imports Infragistics.Olap.FlatData
Imports System
Imports System.Reflection
Imports System.Reflection.Emit

Public Class Form1

#Region " Declarations & FrmCustomize "

    Public FirstSelectedCtl As Control
    Dim ds As FlatDataSource = Nothing
    Dim Rows As New List(Of String)
    Dim Cols As New List(Of String)
    Dim Measures As New List(Of String)
    Dim Filters As New List(Of String)
    Dim Rpt As Integer? = Nothing

#End Region

#Region " Base Form Events "

    Private Sub frmReporteador_Load(sender As Object, e As EventArgs) Handles Me.Load
        'Data Initialization
        LoadCube()
    End Sub

#End Region

#Region " Inicializar Datos de FlatDataSource "

    Private Sub LoadCube()
        Try
            Dim ds As FlatDataSource = InitializePivotData()

            If Not IsNothing(ds) Then
                ulpvgPivot.DataSource = ds
                olapdsSelector.DataSource = ds
                ulpvgPivot.RowHeaderLayout = Infragistics.Win.UltraWinPivotGrid.RowHeaderLayout.Compact
            End If
        Finally

        End Try
    End Sub

    Private Function InitializePivotData() As FlatDataSource
        Try
            ' Initialize the data
            Dim itemSource As IEnumerable(Of Object) = InitData()
            If IsNothing(itemSource) Then Return Nothing

            'Preparar los Measures, Filters y Hierarchies (Vacios si no hay cubo guardado)
            Rows.Clear()
            Cols.Clear()
            Measures.Clear()
            Filters.Clear()

            ' Use "FlatDataSourceInitialSettings" object to initialize Hierarchies and Measures
            Dim settings = New FlatDataSourceInitialSettings() With { _
                .Rows = String.Join(",", Rows), _
                .Columns = String.Join(",", Cols), _
                .Measures = String.Join(",", Measures), _
                .Filters = String.Join(",", Filters) _
            }

            ' Create an instance of the FlatDataSource
            If itemSource.Count = 0 Then Return Nothing
            ds = New FlatDataSource(itemSource, itemSource(0).GetType, settings)

            ' Initialize Measure Descriptor
            AddHandler ds.InitializeMeasureDescriptor, AddressOf ds_InitializeMeasureDescriptor

            ' Initialize Hierarchy Descriptor
            AddHandler ds.InitializeHierarchyDescriptor, AddressOf ds_InitializeHierarchyDescriptor

            ' Events to handle Hierarchies, Measures and Filter changes
            AddHandler ds.AxisAddHierarchyAsyncCompleted, AddressOf ds_AxisAddHierarchyAsyncCompleted
            AddHandler ds.AxisRemoveHierarchyAsyncCompleted, AddressOf ds_AxisRemoveHierarchyAsyncCompleted
            AddHandler ds.AddFilterAsyncCompleted, AddressOf ds_AddFilterAsyncCompleted
            AddHandler ds.RemoveFilterAsyncCompleted, AddressOf ds_RemoveFilterAsyncCompleted
            AddHandler ds.AddMeasureAsyncCompleted, AddressOf ds_AddMeasureAsyncCompleted
            AddHandler ds.RemoveMeasureAsyncCompleted, AddressOf ds_RemoveMeasureAsyncCompleted

            ' Generate cubes and call "InitializeAsync" method passing the control the data is bound to.
            Dim parameters = New CubeGenerationParameters()
            ds.GenerateCube(parameters)
            ds.InitializeAsync(ulpvgPivot)
        Catch ex As Exception
            MessageBox.Show("Ocurrio un error al preparar los datos. Verifique los datos e intentelo nuevamente.")
        Finally

        End Try

        Return ds
    End Function

    Private Sub ds_InitializeHierarchyDescriptor(sender As Object, e As InitializeHierarchyDescriptorEventArgs)
        Dim level As LevelDescriptor = Nothing
        If e.HierarchyDescriptor.LevelDescriptors.TryGetItem("all", level) Then
            level.MemberProvider = Function(item) "(Todos) " & e.HierarchyDescriptor.PropertyDescriptor.Name
        End If
    End Sub

    Private Sub ds_InitializeMeasureDescriptor(sender As Object, e As InitializeMeasureDescriptorEventArgs)
        e.MeasureDescriptor.DisplayFormat = "#,###,##0.00"
        e.MeasureDescriptor.Aggregation = MeasureAggregation.Sum
    End Sub

    Private Sub ds_AxisAddHierarchyAsyncCompleted(ByVal sender As Object, ByVal e As AxisAddHierarchyAsyncCompletedEventArgs)
        If e.AxisType = AxisType.Column Then
            If Not Cols.Contains(e.HierarchyUniqueName) Then Cols.Add(e.HierarchyUniqueName)
        ElseIf e.AxisType = AxisType.Row Then
            If Not Rows.Contains(e.HierarchyUniqueName) Then Rows.Add(e.HierarchyUniqueName)
        End If
    End Sub

    Private Sub ds_AxisRemoveHierarchyAsyncCompleted(ByVal sender As Object, ByVal e As AxisRemoveHierarchyAsyncCompletedEventArgs)
        If e.AxisType = AxisType.Column Then
            If Cols.Contains(e.HierarchyUniqueName) Then Cols.Remove(e.HierarchyUniqueName)
        ElseIf e.AxisType = AxisType.Row Then
            If Rows.Contains(e.HierarchyUniqueName) Then Rows.Remove(e.HierarchyUniqueName)
        End If
    End Sub

    Private Sub ds_AddFilterAsyncCompleted(ByVal sender As Object, ByVal e As AddFilterAsyncCompletedEventArgs)
        If Not Filters.Contains(e.HierarchyUniqueName) Then Filters.Add(e.HierarchyUniqueName)
    End Sub

    Private Sub ds_RemoveFilterAsyncCompleted(ByVal sender As Object, ByVal e As RemoveFilterAsyncCompletedEventArgs)
        If Filters.Contains(e.HierarchyUniqueName) Then Filters.Remove(e.HierarchyUniqueName)
    End Sub

    Private Sub ds_AddMeasureAsyncCompleted(ByVal sender As Object, ByVal e As AddMeasureAsyncCompletedEventArgs)
        If Not Measures.Contains(e.MeasureUniqueName) Then Measures.Add(e.MeasureUniqueName)
    End Sub

    Private Sub ds_RemoveMeasureAsyncCompleted(ByVal sender As Object, ByVal e As RemoveMeasureAsyncCompletedEventArgs)
        If Measures.Contains(e.MeasureUniqueName) Then Measures.Remove(e.MeasureUniqueName)
    End Sub

    Private Function InitData() As IEnumerable(Of Object)
        Dim dt As New DataTable("dt")

        Try
            dt.ReadXmlSchema("dt_schema.xml")
            dt.ReadXml("dt.xml")
        Catch ex As Exception
            MessageBox.Show("Ocurrio un error al ejecutar la consulta. Verifique los datos e intentelo de nuevo.")
            Return Nothing
        End Try

        Dim DynamicData As Type = DynamicDataType.CompileResultType(dt, "Ventas")
        Dim items = New List(Of Object)()

        For Each Row As System.Data.DataRow In dt.Rows
            Dim DynamicDataObject As Object = Activator.CreateInstance(DynamicData)
            DynamicDataType.FillProperties(DynamicDataObject, Row)
            items.Add(DynamicDataObject)
        Next

        Return items
    End Function

#End Region

End Class

Public NotInheritable Class DynamicDataType

    Private Sub New()
    End Sub

    Public Shared Function CompileResultType(dt As DataTable, TypeName As String) As Type
        Dim tb As TypeBuilder = GetTypeBuilder(TypeName)
        Dim constructor As ConstructorBuilder = tb.DefineDefaultConstructor(MethodAttributes.[Public] Or MethodAttributes.SpecialName Or MethodAttributes.RTSpecialName)

        For Each column As DataColumn In dt.Columns
            CreateProperty(tb, column.ColumnName, column.DataType)
        Next

        Dim objectType As Type = tb.CreateType()
        Return objectType
    End Function

    Private Shared Function GetTypeBuilder(TypeName As String) As TypeBuilder
        Dim typeSignature = TypeName
        Dim an = New AssemblyName(typeSignature)
        Dim assemblyBuilder As AssemblyBuilder = AppDomain.CurrentDomain.DefineDynamicAssembly(an, AssemblyBuilderAccess.Run)
        Dim moduleBuilder As ModuleBuilder = assemblyBuilder.DefineDynamicModule("MainModule")
        Dim tb As TypeBuilder = moduleBuilder.DefineType(typeSignature, TypeAttributes.[Public] Or TypeAttributes.[Class] Or TypeAttributes.AutoClass Or TypeAttributes.AnsiClass Or TypeAttributes.BeforeFieldInit Or TypeAttributes.AutoLayout, Nothing)
        Return tb
    End Function

    Private Shared Sub CreateProperty(tb As TypeBuilder, propertyName As String, propertyType As Type)
        Dim fieldBuilder As FieldBuilder = tb.DefineField(Convert.ToString("_") & propertyName, propertyType, FieldAttributes.[Private])

        Dim propertyBuilder As PropertyBuilder = tb.DefineProperty(propertyName, PropertyAttributes.HasDefault, propertyType, Nothing)
        Dim getPropMthdBldr As MethodBuilder = tb.DefineMethod(Convert.ToString("get_") & propertyName, MethodAttributes.[Public] Or MethodAttributes.SpecialName Or MethodAttributes.HideBySig, propertyType, Type.EmptyTypes)
        Dim getIl As ILGenerator = getPropMthdBldr.GetILGenerator()

        getIl.Emit(OpCodes.Ldarg_0)
        getIl.Emit(OpCodes.Ldfld, fieldBuilder)
        getIl.Emit(OpCodes.Ret)

        Dim setPropMthdBldr As MethodBuilder = tb.DefineMethod(Convert.ToString("set_") & propertyName, MethodAttributes.[Public] Or MethodAttributes.SpecialName Or MethodAttributes.HideBySig, Nothing, New Type() {propertyType})

        Dim setIl As ILGenerator = setPropMthdBldr.GetILGenerator()
        Dim modifyProperty As Label = setIl.DefineLabel()
        Dim exitSet As Label = setIl.DefineLabel()

        setIl.MarkLabel(modifyProperty)
        setIl.Emit(OpCodes.Ldarg_0)
        setIl.Emit(OpCodes.Ldarg_1)
        setIl.Emit(OpCodes.Stfld, fieldBuilder)

        setIl.Emit(OpCodes.Nop)
        setIl.MarkLabel(exitSet)
        setIl.Emit(OpCodes.Ret)

        propertyBuilder.SetGetMethod(getPropMthdBldr)
        propertyBuilder.SetSetMethod(setPropMthdBldr)
    End Sub

    Public Shared Sub FillProperties(SenderObject As Object, Row As DataRow)
        For Each Column As DataColumn In Row.Table.Columns
            Dim ObjectProperty As PropertyInfo = SenderObject.GetType.GetProperty(Column.ColumnName)

            If IsDBNothing(Row(Column.ColumnName)) AndAlso Column.DataType Is GetType(String) Then
                ObjectProperty.SetValue(SenderObject, IsDBNothing(Row(Column.ColumnName), Column.ColumnName & "- N/A"), Nothing)
            Else
                ObjectProperty.SetValue(SenderObject, IsDBNothing(Row(Column.ColumnName), Nothing), Nothing)
            End If
        Next
    End Sub

    Public Shared Function IsDBNothing(ByVal ValueToTest As Object) As Boolean
        Return IsNothing(ValueToTest) OrElse IsDBNull(ValueToTest)
    End Function

    Public Shared Function IsDBNothing(ByVal ValueToTest As Object, ByVal RetValIfIsDBNothing As Object) As Object
        If IsDBNothing(ValueToTest) Then
            Return RetValIfIsDBNothing
        Else
            Return ValueToTest
        End If
    End Function

End Class